""" Task Router

The primary function is it to route Tasks between the front and Backend socket end points and
provide Service Discovery of provided Worker functions. The seperate Sockets allow to bridge
internal / external networks.

Example Usage:

router = butler.Router()
router.frontend.bind("tcp://*:5555")
router.backend.bind("tcp://*:5556")
router.run()
"""

import time
import datetime
import gevent
import logging
import zmq.green as zmq

HEARTBEAT_INTERVAL = 1000  # msec between heartbeats sent out
HEARTBEAT_LIVENESS = 3  # seconds until heartbeat is expected / worker is considered dead
REQUEST_LIFESPAN = 1  # seconds

PPP_READY = "\x01"  # Signals worker is ready
PPP_HEARTBEAT = "\x02"  # Signals worker heartbeat
PPP_RECONNECT = "\x03"  # Signals worker re-connect
PPP_REPLY = "\x04"  # Signals worker message

log = logging.getLogger(__name__)


class Service(object):
    """a single Service"""

    name = None  # Service name
    requests = None  # List of client requests
    waiting = None  # List of waiting workers

    def __init__(self, name):
        self.updated_at = datetime.datetime.now()
        self.name = name
        self.requests = []
        self.waiting = []

    def __repr__(self):
        return '<Service - %s>' % self.name


class Worker(object):
    """an idle or active Worker instance"""

    address = None  # routing address
    expiry = None  # expires at this point - unless heartbeat
    service = None  # owning service name if known

    def __init__(self, address):
        self.address = address
        self.expiry = time.time() + HEARTBEAT_LIVENESS

    def __repr__(self):
        return '<Worker - %s>' % self.address


class EBRouter(object):

    heart = None

    def __init__(self):

        self.services = {}  # `Worker` objects grouped by Service to allow for faster discovery

        self.workers = {}  # index of all `Worker` objects by address

        self.waiting = []  # direct access to check heartbeat status of Workers

        self.heartbeat_at = time.time() + 1e-3 * HEARTBEAT_INTERVAL

        self.context = zmq.Context(1)

        self.frontend = self.context.socket(zmq.ROUTER)  # Front End
        self.frontend.setsockopt(zmq.HWM, 0)

        self.backend = self.context.socket(zmq.ROUTER)  # Back End
        self.backend.setsockopt(zmq.HWM, 0)

        self.pull_backends = zmq.Poller()
        self.pull_backends.register(self.backend, zmq.POLLIN)

        self.poll_both = zmq.Poller()
        self.poll_both.register(self.frontend, zmq.POLLIN)
        self.poll_both.register(self.backend, zmq.POLLIN)

    def setup_heartbeat(self):

        log.debug('Setup Heartbeat')

        heartbeat_socket = self.context.socket(zmq.PUSH)
        heartbeat_socket.connect('tcp://localhost:5556')

        while True:

            if time.time() < self.heartbeat_at:
                gevent.sleep(0.1)
                continue

            log.debug('attempting to ping %i worker' % len(self.waiting))

            self.purge_workers()

            for worker in self.waiting:

                msg = [PPP_REPLY, worker.address, PPP_HEARTBEAT]
                heartbeat_socket.send_multipart(msg)

            self.heartbeat_at = time.time() + 1e-3 * HEARTBEAT_INTERVAL

    def poll_sockets(self):

        while True:

            # ignore FE / client requests if no workers are connected to the backend (no PPP_READY received)
            if not len(self.workers):
                log.warn('No Worker available - Waiting for Worker to join on BE')
                poller = self.pull_backends
            else:
                poller = self.poll_both

            gevent.sleep(0)

            # poll socket for request message up to HEARTBEAT_INTERVAL seconds
            socks = dict(poller.poll(HEARTBEAT_INTERVAL))

            # handle worker activity on backend
            if self.backend in socks and socks[self.backend] == zmq.POLLIN:

                log.debug('worker BE activity')
                self.process_worker()

            # Client request received - forward it to the backend router
            if self.frontend in socks and socks[self.frontend] == zmq.POLLIN:

                log.debug('client FE activity')
                self.process_client()

    def run(self):

        # setup router Heartbeat
        gevent.spawn(self.setup_heartbeat)

        gevent.spawn(self.poll_sockets).join()

    def process_worker(self):
        """Process Service Worker Message"""

        # read request from BE
        frames = self.backend.recv_multipart()

        assert len(frames) >= 1

        log.debug('process worker msg: %s' % frames)

        # parse response
        worker_uuid = frames[0]
        msg = frames[1:]

        # worker control message received
        if not msg:
            log.debug('empty response')
            return

        command = msg.pop(0)

        # fires via `Worker.__init__`
        if command == PPP_READY:

            # worker had been registered already / active session open
            worker_registered = worker_uuid in self.workers

            # not first command after startup
            if worker_registered:
                worker = self.workers[worker_uuid]
                log.critical('Late PPP_READY received - de-register Worker: %s' % worker)
                self.delete_worker(worker, disconnect = True)
                return

            service = msg.pop(0)

            log.debug('PPP_READY received from "%s" worker: %s' % (
                service, worker_uuid
            ))

            # create `Worker` object
            worker = self.require_worker(worker_uuid)

            # attach worker to service and mark as idle
            worker.service = self.require_service(service)

            self.worker_waiting(worker)

        elif command == PPP_HEARTBEAT:

            # worker had been registered already / active session open
            worker_registered = worker_uuid in self.workers

            log.debug('PPP_HEARTBEAT received from worker: %s' % worker_uuid)

            if not worker_registered:

                log.warn('PPP_REPLY received but Worker not registered yet: %s' % worker_uuid)
                self.disconnect_worker( worker_uuid )

            else:

                worker = self.workers[worker_uuid]

                worker.expiry = time.time() + HEARTBEAT_LIVENESS

        elif command == PPP_REPLY:

            # worker had been registered already / active session open
            worker_registered = worker_uuid in self.workers

            # PPP_REPLY came from worker server
            if worker_registered:

                worker = self.workers[worker_uuid]

                log.debug('Worker PPP_REPLY received from %s - forward to Client %s' % (
                    worker_uuid, msg
                ))

                self.frontend.send_multipart(msg)

                self.worker_waiting(worker)

            # PPP_REPLY should go to a registered Worker server
            elif frames[2] in self.workers:

                worker = self.workers[frames[2]]

                log.debug('Route PPP_REPLY to Worker: %s' % msg)

                self.backend.send_multipart(msg)

            else:
                log.warn('Unknown PPP_REPLY received - attempting: %s' % msg)

                # send message to both FE and BE
                self.frontend.send_multipart(msg)
                self.backend.send_multipart(msg)

        elif msg:

            log.info('Unknown Backend Request received: %s' % msg)

    def process_client(self):

        # read request as multi part message from FE router socket
        frames = self.frontend.recv_multipart()

        assert len(frames) >= 2  # service name + request body

        log.debug('process client request: %s' % frames)

        # parse request
        ident, null, request_id, service_name, request_expiration, request = frames

        # discared if request is expired
        if int(request_expiration) < int(round((time.time() * 1000))):
            log.warn('request expired at: %s' % request_expiration)
            return

        if service_name.startswith('mmi.'):

            log.debug('internal service call')
            self.service_lookup(frames)

        else:
            log.info('new request: %s | %s' % (service_name, request_id))

            self.dispatch_request(
                self.require_service(service_name), frames
            )

    def dispatch_request(self, service, msg):

        assert (service is not None)

        self.purge_workers()

        if service.waiting:
            log.debug("dispatch request: %s" % service.waiting)
        else:
            log.warn('no Worker available for "%s" - discarding request' % service)
            empty_msg = msg[:2] + ['404', ]
            return self.frontend.send_multipart(empty_msg)

        if msg is not None:  # queue message in memory
            service.requests.append(msg)

        while service.waiting and service.requests:

            msg = service.requests.pop(0)

            new_worker = service.waiting.pop(0)

            log.debug('forwarding Client request Worker BE: %s' % new_worker)

            # add the Worker address to the request to route the message
            msg.insert(0, new_worker.address)

            # send message to backend
            self.backend.send_multipart(msg)

    def service_lookup(self, msg):

        service_name = msg[3].replace('mmi.', '')

        returncode = '501'
        returncode = '200' if service_name in self.services else '400'

        msg = msg[:2] + [service_name, returncode]

        # send response to client Front End router
        self.frontend.send_multipart(msg)

    def require_service(self, name):
        """ Locates the service (registers if necessary)
        """
        assert (name is not None)
        service = self.services.get(name)

        if not service:
            log.info('registering new Service: %s' % name)
            service = Service(name)
            self.services[name] = service
        return service

    def require_worker(self, worker_uuid):
        """ Find or create `Worker` record
        """
        assert (worker_uuid is not None)
        worker = self.workers.get(worker_uuid)

        if not worker:  # create new `Worker`
            worker = Worker(worker_uuid)
            self.workers[worker_uuid] = worker

            log.info("registering new Worker: %s | %s" % ( worker.address, worker ))

        return worker

    def worker_waiting(self, worker):
        """This worker is now waiting for work.
        """
        log.debug('worker ready for work call: %s | %i' % (
                worker, len(self.waiting)
            ))

        # Queue to router and service waiting lists
        if worker not in self.waiting:
            self.waiting.append(worker)

        if worker not in worker.service.waiting:
            worker.service.updated_at = datetime.datetime.now()
            worker.service.waiting.append(worker)

        worker.expiry = time.time() + HEARTBEAT_LIVENESS

    def purge_workers(self):
        """ Look for & kill expired workers.
        Workers are oldest to most recent.
        """

        for worker in self.waiting:
            if worker.expiry < time.time():
                log.warning("deleting expired worker: %s" % worker.address)
                self.waiting.remove(worker)
                self.delete_worker(worker, False)

        for service_name in self.services.keys():
            if not self.services[service_name].waiting:
                if self.services[service_name].updated_at <= (
                    datetime.datetime.now() - datetime.timedelta(hours = 1)
                ):
                    log.warn('deleting expired Service: %s' % service_name)
                    self.services.pop(service_name)

    def disconnect_worker(self, worker_address):

        log.warn('sent Worker PPP_RECONNECT to: %s' % worker_address)

        control_socket = self.context.socket(zmq.PUSH)
        control_socket.connect('tcp://localhost:5556')
        msg = [PPP_REPLY, worker_address, PPP_RECONNECT]
        control_socket.send_multipart(msg)
        control_socket.close()

    def delete_worker(self, worker, disconnect = False):
        """Deletes worker from all data structures, and deletes worker."""

        assert worker is not None

        if worker.service is not None:
            if worker in worker.service.waiting:
                worker.service.waiting.remove(worker)

        while True:
            if worker in self.waiting:
                self.waiting.remove(worker)
            else: break

        # remove from registry
        self.workers.pop(worker.address)

        if disconnect:  # send re-connect msg to worker
            self.disconnect_worker(worker.address)



"""
Example Usage:

worker = butler.Service("tcp://localhost:5556", 'video.cut')

reply = None
while True:
    request = worker.recv(reply)
    if not request:
        break  # worker was interrupted / stopped
"""

import time
import uuid
import json
import sys
import cPickle
import gevent
import logging
import zmq.green as zmq

HEARTBEAT_INTERVAL = 1  # seconds between a PPP_HEARTBEAT is send to the broker
HEARTBEAT_LIVENESS = 3  # 3 seconds until PPP_HEARTBEAT is expected from broker or considered dead

INTERVAL_INIT = 0
INTERVAL_MAX = 32

PPP_READY = "\x01"  # Signals worker is ready
PPP_HEARTBEAT = "\x02"  # Signals worker heartbeat
PPP_RECONNECT = "\x03"  # Signals worker re-connect
PPP_REPLY = "\x04"  # Signals worker reply

log = logging.getLogger(__name__)


class EBWorker(object):

    broker = None
    context = None
    poller = None
    service = None
    worker = None

    liveness = HEARTBEAT_LIVENESS
    interval = INTERVAL_INIT

    expect_reply = False
    reply_to = None

    heartbeat_at = None

    heart = None

    sink = None

    def __init__(self, bind_address, service = 'echo'):

        self.context = zmq.Context(1)

        self.poller = zmq.Poller()

        self.service = service

        self.bind_address = bind_address

        self.uuid = str(uuid.uuid4())

        self.rpc_registry = {}

        # PUSH socket to send broadcast/flow messages to
        self.sink = self.context.socket(zmq.PUSH)
        self.sink.connect('tcp://localhost:5558')

        # DEALER socket to get jobs from/to
        self.setup_worker_socket()

        # init heartbeat
        self.heartbeat_at = time.time() + HEARTBEAT_INTERVAL

    def setup_worker_socket(self):
        """Helper function that returns a new configured socket
           connected to the queue"""

        # close existing socket
        if self.worker:
            log.warn('re-connect socket')
            self.poller.unregister(self.worker)
            self.worker.close()

        # create DEALER socket
        self.worker = self.context.socket(zmq.DEALER)
        self.worker.setsockopt(zmq.HWM, 0)

        self.worker.setsockopt(zmq.IDENTITY, self.uuid)

        # register worker socket with poller
        self.poller.register(self.worker, zmq.POLLIN)

        # connec to ROUTER socket
        self.worker.connect(self.bind_address)

        # send `PPP_READY` message to Router
        self.signal_ready()

    def setup_heartbeat(self):

        log.info('setup_heartbeat with %s' % self.bind_address)

        time_run = 0

        heartbeat_socket = self.context.socket(zmq.PUSH)
        heartbeat_socket.connect(self.bind_address)

        while True:

            # sync the time between pings due the GIL
            last_ping = time.time() - time_run
            if last_ping < HEARTBEAT_INTERVAL:
                time_sleep = HEARTBEAT_INTERVAL - last_ping
            else:
                time_sleep = HEARTBEAT_INTERVAL
            time_run = time.time()
            gevent.sleep(time_sleep)

            # send the msg together with Worker UUID to Router
            heartbeat_socket.send_multipart(
                    [ PPP_HEARTBEAT, self.uuid ]
                )

            self.heartbeat_at = time.time() + HEARTBEAT_INTERVAL

            log.debug('Worker heartbeat SENT')

    def signal_heartbeat(self):

        if time.time() > self.heartbeat_at:

            # send the msg together with Worker UUID to Router
            self.worker.send_multipart(
                    [ PPP_HEARTBEAT ]
                )

            self.heartbeat_at = time.time() + HEARTBEAT_INTERVAL

            log.debug('signal_heartbeat - SENT')
        else:
            log.debug('signal_heartbeat - NOT DUE yet (%s)' % (
                    time.time() - self.heartbeat_at
                ))

    def signal_ready(self):
        """ Signals that the Worker is ready to accept work (async job load) """
        # send `PPP_READY` message to Router
        log.info('sent PPP_READY - register: %s | %s' % (self.service, self.uuid))

        self.worker.send_multipart(
                [ PPP_READY, self.service ]
            )

    def send(self, message):
        """ Send replies to the Client """
        assert self.reply_to is not None

        #if isinstance(message, str):
        #    message = [self.reply_to, message, ]

        msg = [PPP_REPLY, self.reply_to, '', self.api_call, message]

        log.debug('sending reply: %s' % msg)
        self.worker.send_multipart(msg)

    def recv(self, reply=None):
        """Send reply, if any, to broker and wait for next request."""

        if reply is not None:
            self.send(reply)

        gevent.sleep(0)

        # poll broker socket - expecting a reply within HEARTBEAT_INTERVAL seconds
        socks = dict(self.poller.poll(HEARTBEAT_INTERVAL * 1000))

        # Handle worker activity on backend
        if socks.get(self.worker) == zmq.POLLIN:

            frames = self.worker.recv_multipart()

            log.debug('FE request')

            # reset heartbeat timeout
            self.liveness = HEARTBEAT_LIVENESS
            self.interval = INTERVAL_INIT

            if frames and frames[0] == PPP_HEARTBEAT:

                log.debug('Queue heartbeat RECEIVED')

            elif frames and frames[0] == PPP_RECONNECT:

                log.warning("Queue re-connect RECEIVED")

                self.reconnect_broker()

            elif len(frames) == 6:

                # parse client request
                client_ident, x, api_call, service_name, expiration, request = frames

                log.debug('New Request: %s' % frames)

                self.reply_to = client_ident
                self.api_call = api_call

                return request

            else:
                log.critical('Invalid Request: %s' % frames)

        else:  # no response received from router socket
            log.debug('no response from Router')

            self.liveness -= 1
            if self.liveness <= 1:

                log.warn("Heartbeat DEAD (%i seconds) - Reconnecting to Router in %0.2fs" % (
                    self.liveness, self.interval
                ))
                gevent.sleep(self.interval)

                if not self.interval:
                    self.interval = 1
                elif self.interval < INTERVAL_MAX:
                    self.interval *= 2
                else:
                    self.interval = INTERVAL_INIT

                self.reconnect_broker()

        self.signal_heartbeat()

    def reconnect_broker(self):

        # create new socket to broker
        self.setup_worker_socket()

        # reset heartbeat timeout
        self.liveness = HEARTBEAT_LIVENESS

    def register(self, object_callback):
        """Register a python object for RPC use"""

        log.debug('RPC register object: %s' % object_callback)
        self.rpc_registry['.'] = object_callback

    def register_function(self, function_callback, function_name = None):
        """ registers a python RPC function """

        if not function_name:
            function_name = function_callback.func_name

        log.debug('RPC register function: %s as "%s"' % (
            function_callback, function_name
        ))
        self.rpc_registry[function_name] = function_callback

    def run(self):
        reply = None

        while True:
            log.debug('polling for work (reply: %s)' % reply)

            request = self.recv(reply)
            reply = None  # reset response

            if not request:
                log.debug('empty `request` received')
                continue

            log.debug('got RPC request to process: %s' % request)

            # accepts JSON requests
            rpc_request = json.loads(request)

            # single methods have been registered
            if rpc_request['method'] in self.rpc_registry:
                try:
                    reply = self.rpc_registry[rpc_request['method']](
                            *rpc_request['args'],
                            **rpc_request['kwargs']
                        )
                except:
                    reply = '500:exception:%s' % cPickle.dumps(
                        sys.exc_info()[1]
                    )

            # object has been registered for callbacks
            elif '.' in self.rpc_registry:
                try:
                    reply = getattr(self.rpc_registry['.'], rpc_request['method'])(
                            *rpc_request['args'],
                            **rpc_request['kwargs']
                        )
                except:
                    reply = '500:exception:%s' % cPickle.dumps(
                        sys.exc_info()[1]
                    )
            else:
                reply = '404'



# -*- coding: utf-8 -*-
"""
The `butler` framework aims to make building reliable and high-performance service-oriented
systems easy. The internal communication is done via ØMQ sockets.

Usage
-----

Start a Router::

    import butler

    router = butler.Router()
    router.frontend.bind("tcp://*:5555")
    router.backend.bind("tcp://*:5556")
    router.run()

Register a Service Worker by name::

    worker = butler.Service('tcp://127.0.0.1:5556', 'api.images')

Expose a function on the Service Worker::

    def resize_image(name, size):
        return 'resized image'

    worker.register_function(resize_image)
    worker.run()

Or expose all the methods of an object::

    class RPCService(object):
        def resize_image(self, name, size):
            return 'resized image'

    worker.register(RPCService())
    worker.run()

Send a request to a registered Service and receive its response::

    client = butler.Client('tcp://127.0.0.1:5555').rpc('api.images')
    client.resize_image('test.jpeg', '150x180')

"""

from setuptools import setup

setup(
    name = 'butler',
    version = '0.2.0',
    author = 'Manuel Kreutz',
    author_email = 'manuel@140.am',
    description = '0MQ based Service Oriented RPC Framework',
    license = 'MIT',
    url = 'https://github.com/encodingbooth/butler',
    packages = ['butler'],
    long_description = __doc__,
    classifiers = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Programming Language :: Python :: 2.5',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Clustering',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Networking',
        'Topic :: System :: Systems Administration',
        'License :: OSI Approved :: MIT License',
    ],
    install_requires = [
        'pyzmq==2.2.0.1',
        'gevent'
    ]
)



""" Task Result Sink

Provides a PULL zeromq socket to receive Worker messages/events

Example Usage:

sink = butler.Sink('tcp://*:5558')
while True:
    msg = sink.get_message()
    log.info('MSG received: %s' % msg)

worker = butler.Service("tcp://localhost:5556", 'image.validate')
worker.sink.send('task received')
"""

import logging
import zmq.green as zmq

log = logging.getLogger(__name__)


class EBSink(object):

    context = None
    poller = None
    sink = None

    bind_host = None

    def __init__(self, bind_host = '*:5558'):

        self.context = zmq.Context(1)
        self.poller = zmq.Poller()

        # PULL socket to receive broadcast/flow messages on
        self.sink = self.context.socket(zmq.PULL)

        # setting an identity so worker will buffer outgoing messages in case of downtime
        self.sink.setsockopt(zmq.IDENTITY, 'sink')

        self.sink.bind(bind_host)

        self.setup_polling()

    def setup_polling(self):

        self.poller.register(self.sink, zmq.POLLIN)

    def get_message(self):

        msg = ''

        socks = dict(self.poller.poll())

        if self.sink in socks and socks[self.sink] == zmq.POLLIN:
            msg = self.sink.recv()

        return msg



""" Encoding Booth - Worker Broker Client

A 0MQ socket of type ZMQ_REQ is used by a client to send requests to and receive replies from a service.
Each request sent is round-robined among all broker services and each reply received is matched with the
last issued request.
"""

import time
import logging
import json
import uuid
import cPickle
import butler
import gevent
import zmq.green as zmq

REQUEST_RETRIES = 1
REQUEST_TIMEOUT = 2500

log = logging.getLogger(__name__)


class RPCProxyCall:

    def __init__(self, client_obj, service_name, service_attr):
        self.service = service_name
        self.service_attr = service_attr
        self.client = client_obj

    def __call__(self, *args, **kwargs):
        self.client.retries = 1

        response = self.client.call(
            self.service, {
                'method' : self.service_attr,
                'args' : args,
                'kwargs' : kwargs
            }
        )

        # service returning `404` is considered a non implemented function
        if response == '404':
            raise AttributeError('%r object has no attribute %r' % (
                type(self).__name__, self.service_attr
            ))
        elif response and response.startswith('500:exception:'):
            raise cPickle.loads(response[14:])
        else:
            return response


class RPCProxy(object):

    def __init__(self, client_obj, service_name):
        self.service = service_name
        self.client = client_obj
        self.client.persistent = False

    def __setattr__(self, attr, value):
        # handle setting of `Client.timeout`
        if attr == 'timeout':
            self.client.timeout = value
        else:
            super(RPCProxy, self).__setattr__(attr, value)

    def __getattr__(self, attr):
        return RPCProxyCall(self.client, self.service, attr)


class EBClient(object):

    broker = None  # 0MQ address - example: tcp://127.0.0.1:5555
    context = None  # zmq.Context
    client = None  # zmq.Socket
    poller = None  # zmq.Poller

    retries = REQUEST_RETRIES  # count - number of request retries

    timeout = REQUEST_TIMEOUT * 1e-3

    sequence = 0

    persistent = True

    response = {}

    def __init__(self, bind_address):

        self.uuid = str(uuid.uuid4())

        self.bind_address = bind_address

        self.context = zmq.Context(1)

        self.poller = zmq.Poller()

        self.connect_to_broker()

    def connect_to_broker(self):

        # close existing socket
        if self.client:
            self.poller.unregister(self.client)
            self.client.close()

        # create a ZMQ_REQ socket to send requests / receive replies
        self.client = self.context.socket(zmq.DEALER)
        self.client.setsockopt(zmq.HWM, 0)
        self.client.linger = 1
        self.client.setsockopt(zmq.IDENTITY, self.uuid)

        # register socket with poller
        self.poller.register(self.client, zmq.POLLIN)

        # connect to `Router` socket
        self.client.connect(self.bind_address)

    def call(self, service, request):
        """
        Returns the reply message or None if there was no reply within timeout limit
        The maximum total blocking time for this method is REQUEST_RETRIES*REQUEST_TIMEOUT in msec
        """

        # will hold the response
        reply = None

        # convert python dict request to JSON structure
        if isinstance(request, dict):
            request = json.dumps(request)

        # request expiration time in milliseconds for `call`
        request_expiration = int(
                round(time.time() + (self.timeout * REQUEST_RETRIES)) * 1000
            )

        # unix time at which the polling should end (`request timeout`)
        call_time_end = time.time() + (self.timeout * REQUEST_RETRIES)

        # number of attempts to re-send request in case of no service available
        request_retries = REQUEST_RETRIES

        # attempt to get a response from router up to REQUEST_RETRIES times
        while request_retries > 0:

            """ Request Message Format:

            Frame 0: Empty (zero bytes)
            Frame 1: "EBv1:%i" (string, representing version and unique request ID)
            Frame 2: Service name (printable string)
            Frame 3: Expiration Time (unix time in future)
            Frame 4: Request body (opaque binary, converted to JSON string)
            """

            request_id = str(uuid.uuid4())

            msg = [
                '',
                '%s:%s' % (butler.__version__, request_id),
                str(service),
                str(request_expiration),
                str(request)
            ]

            # in case persistent is disabled re-send and connect on each retry attempt
            if self.persistent and request_retries == REQUEST_RETRIES:
                self.client.send_multipart(msg)
            elif not self.persistent:
                self.client.send_multipart(msg)

            log.debug('sending request [%s | %s]: %s' % (
                service, request_id, json.loads(request)['method']
            ))

            # after dispatching the request poll up to max timeout for response
            response_poll = gevent.spawn(
                self.poll_for_reply, request_id, call_time_end
            )

            # at same time look for the per process stored response
            response_reply = gevent.spawn(
                    self.find_local_reply, request_id, call_time_end, response_poll
                )

            gevent.joinall(
                [ response_poll, response_reply ]
            )

            reply = response_reply.value

            request_retries -= 1

        return reply

    def rpc(self, service_name):

        self.rpc_proxy = RPCProxy(
            self, service_name
        )
        return self.rpc_proxy

    def poll_for_reply(self, orig_request_id, time_expiration):

        while time.time() < time_expiration:

            gevent.sleep(0.001)

            socks = dict(self.poller.poll(1000))

            if socks.get(self.client) == zmq.POLLIN:

                # receive its response
                frames = self.client.recv_multipart()

                if not frames:
                    log.warn('got empty reply back from `Broker`')
                    continue

                if len(frames) > 2:

                    # ensure the msg can be understood / protocol API version
                    if frames[1].startswith(butler.__version__):

                        log.debug('API %s msg received: %s' % (
                            butler.__version__, frames
                        ))

                        # ident, service_name, function, expiration, request_body = frames
                        x, api_call, request_body = frames

                        # take the API version / request sequence apart
                        request_id = api_call.split('%s:' % butler.__version__)[1]

                        self.response[request_id] = request_body

                        if orig_request_id == request_id:
                            break
                    else:
                        log.warn('Invalid API msg received: %s' % frames)
                else:
                    log.warn('raw response received: %s' % frames)
            else:
                log.debug('no response from router')

    def find_local_reply(self, orig_request_id, time_expiration, response_poll):

        while time.time() < time_expiration:
            reply = None
            if orig_request_id in self.response:
                reply = self.response.pop(orig_request_id)
                response_poll.kill()
                break
            gevent.sleep(0.001)

        return reply
